%{
/*-
 * ACPI Source Language Compiler
 *
 * Copyright (c) 2000 Mitsuru IWASAKI <iwasaki@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <sys/param.h>

#include <stdlib.h>
#include <string.h>

#include "aslc.h"
#include "y.tab.h"

#define ACTION_SERIALIZERULE(value)	do {				\
	yylval.num = value;						\
	return(TOK_SERIALIZERULEKEYWORD);				\
} while (0)

#define ACTION_VARIABLE(pos, token)	do {				\
	yylval.num = strtol(&yytext[pos], NULL, 10);			\
	return(token);							\
} while (0)

#define ACTION_INTEGER(base)	do {					\
	yylval.lit.type = LITERAL_INTEGER;				\
	yylval.lit.num.val = strtol(yytext, NULL, base);		\
	if (yylval.lit.num.val < 0xff) {				\
		return(TOK_BYTECONST);					\
	}								\
	if (yylval.lit.num.val < 0xffff) {				\
		return(TOK_WORDCONST);					\
	}								\
	return(TOK_DWORDCONST);						\
} while (0)

int		parser_lineno;

static int	comment_start = 1;

%}

%x CODE
%x COMMENT

%%

%{
	if (parser_not_initialized) {
		BEGIN(CODE);
		parser_lineno = 1;
		parser_not_initialized = 0;
	}
%}

 /*
  * Newline
  */

<CODE,COMMENT>\n	{ parser_lineno++; }

 /*
  * Comment
  */

 /* one-line omment */
<CODE>"//".*$		;

 /* normal style comment */
<CODE>"/*"		{ comment_start = parser_lineno; BEGIN(COMMENT); }
<COMMENT>[^*\n]*	;	/* eat anything that's not a '*' */
<COMMENT>"*"+[^*/\n]*	;	/* eat up '*'s not followed by '/'s */
<COMMENT>"*"+"/"	{ BEGIN(CODE); }
<COMMENT><<EOF>>	{
				parser_error("unterminated comment "
				    "beginning from line %d\n", comment_start);
				 yyterminate();
			}

 /*
  * ASL code
  */

<CODE>[ \t]+		;
<CODE>[,{}()]		{ return(yytext[0]); }

<CODE>DefinitionBlock	{ return(TOK_DEFINITIONBLOCK); }

 /* Named Object */
<CODE>CreateByteField	{ return(TOK_CREATEBYTEFIELD); }
<CODE>Method		{ return(TOK_METHOD); }

 /* Type1Opcode */
<CODE>While		{ return(TOK_WHILE); }

 /* Type2Opcode */
<CODE>Decrement		{ return(TOK_DECREMENT); }
<CODE>Index		{ return(TOK_INDEX); }
<CODE>RefOf		{ return(TOK_REFOF); }
<CODE>Store		{ return(TOK_STORE); }

 /* Keyword */
<CODE>Serialized	{ ACTION_SERIALIZERULE(1); }
<CODE>NotSerialized	{ ACTION_SERIALIZERULE(0); }

 /* SuperName */
<CODE>Arg[0-6]		{ ACTION_VARIABLE(3, TOK_ARGTERM); }
<CODE>Local[0-7]	{ ACTION_VARIABLE(5, TOK_LOCALTERM); }
<CODE>Debug		{ return(TOK_DEBUGTERM); }

 /* Integer */
<CODE>0x[0-9a-fA-F]+	{ ACTION_INTEGER(16); }
<CODE>[0-9]+		{ ACTION_INTEGER(10); }

 /* String */
<CODE>\"[^"]*\"		{
				size_t	i, len;

				len = yyleng - 2;
				yylval.lit.type = LITERAL_STRING;
				yylval.lit.str.ptr = (char *)malloc(len + 1);
				if (yylval.lit.str.ptr == NULL) {
					parser_fatal("No more room for String");
				}
				strncpy(yylval.lit.str.ptr, &yytext[1], len);
				yylval.lit.str.ptr[len] = '\0';
				for (i = 0; i < len; i++) {
					if (yylval.lit.str.ptr[i] == '\n') {
						parser_lineno++;
					}
				}
				return(TOK_STRING);
			}

 /* DataObject */
<CODE>Buffer		{ return(TOK_BUFFER); }
<CODE>Package		{ return(TOK_PACKAGE); }

 /* ASL Names */
<CODE>[\\.^]		{ return(yytext[0]); }
<CODE>[A-Z_][A-Z_0-9]{0,3}	{
				yylval.str = strdup(yytext);
				return(TOK_NAMESEG);
			}
 /* Error */
<CODE>[A-Z_][A-Z_0-9]{4,}	{
				parser_error("<%s> is too long", yytext);
				return(TOK_ERROR);
}
<CODE>[^/*\n \t,{}()0-9a-fA-Z".^_]+	{
				parser_error("<%s> is unexpected", yytext);
				return(TOK_ERROR);
			}
%%
