%{
/*-
 * ACPI Source Language Compiler
 *
 * Copyright (c) 2000 Mitsuru IWASAKI <iwasaki@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <sys/param.h>

#include <err.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "aslc.h"

#ifdef DEBUG
#define YYDEBUG 1
#define PARSER_DEBUGPRINT(phase, x...)	do {				\
	if (parser_phase <= phase && parser_debuglevel) {		\
		printf(x);						\
	}								\
} while (0)

#else
#define PARSER_DEBUGPRINT(phase, x...)
#endif

#define PARSER_ENCODING_TERM(code, term...)	do {			\
	if (parser_phase == 1) {					\
		current_state->pkglength++;				\
	}								\
	snprintf(parser_msgbuf, PARSER_MSGBUFSIZE, term);		\
	PARSER_DEBUGPRINT(2, "0x%02x:\t%s:\tat line %d\n",		\
	    code, parser_msgbuf, parser_lineno);			\
} while (0)

#define PARSER_ENCODING_PREFIX(val, type)	do {			\
	int	prefix;							\
	if (parser_phase == 1) {					\
		current_state->pkglength++;				\
	}								\
	if (type == LITERAL_INTEGER) {					\
		if (val < 0xff) {					\
			prefix = 0x0a;					\
		} else if (val < 0xffff) {				\
			prefix = 0x0b;					\
		} else {						\
			prefix = 0x0c;					\
		}							\
	} else {							\
		prefix = 0x0d;						\
	}								\
	PARSER_DEBUGPRINT(2, "0x%02x:\tPrefix:\tat line %d\n",		\
	    prefix, parser_lineno);					\
} while (0)

#define PARSER_ENCODING_INTEGER(val, bytes, term)	do {		\
	if (current_state->prefix_expected) {				\
		PARSER_ENCODING_PREFIX(val, LITERAL_INTEGER);		\
	}								\
	switch (bytes) {						\
	case 1:								\
		PARSER_ENCODING_TERM(val, term);			\
		break;							\
	case 2:								\
		PARSER_ENCODING_TERM((val >> 8) & 0xff, term);		\
		PARSER_ENCODING_TERM((val & 0xff), term);		\
		break;							\
	case 4:								\
		PARSER_ENCODING_TERM((val >> 24) & 0xff, term);		\
		PARSER_ENCODING_TERM((val >> 16) & 0xff, term);		\
		PARSER_ENCODING_TERM((val >> 8) & 0xff, term);		\
		PARSER_ENCODING_TERM((val & 0xff), term);		\
		break;							\
	default:							\
		break;							\
	}								\
} while (0)

#define PARSER_BEGINPACKAGE(x...)	do {				\
	PARSER_DEBUGPRINT(2, "\t<==\t");				\
	PARSER_DEBUGPRINT(2, x);					\
	PARSER_DEBUGPRINT(2, "\n");					\
} while (0)

#define PARSER_ENDPACKAGE(term)	do {					\
	u_int8_t	code[4];					\
	int		len, i;						\
									\
	parser_pkglength_fixup();					\
	len = current_state->pkglength;					\
	len = parser_pkglength_encoding(len, code);			\
	for (i = 0; i < len; i++) {					\
		PARSER_DEBUGPRINT(2, "\t===>\t%s:0x%x<PkgLength>\n",	\
			    term, code[i]);				\
	}								\
	parser_state_pop();						\
} while (0)

#define PARSER_PKGLENGTH_FLUSH()	do {				\
	u_int8_t	code[4];					\
	int		len, i;						\
									\
	len = current_state->pkglength;					\
	len = parser_pkglength_encoding(len, code);			\
	for (i = 0; i < len; i++) {					\
		PARSER_ENCODING_TERM(code[i], "PkgLength");		\
	}								\
	if (parser_phase == 1) {					\
		current_state->pkglength -= len;			\
	}								\
} while (0)

#define PARSER_SYMTAB_NEW(dest, name, type)	do {			\
	dest = symtab_find((const char *)name);				\
	if (dest != NULL) {						\
		snprintf(parser_msgbuf, PARSER_MSGBUFSIZE,		\
		    "<%s> is already defined", name);			\
		yyerror(parser_msgbuf);					\
		exit(1);						\
	} else {							\
		dest = symtab_new((const char *)name, type);		\
	}								\
} while (0)

#define PARSER_SYMTAB_FIND(dest, name, typestr)	do {			\
	dest = symtab_find((const char *)name);				\
	if (dest == NULL) {						\
		snprintf(parser_msgbuf, PARSER_MSGBUFSIZE,		\
		    "%s<%s> is not defined", typestr, name);		\
		yyerror(parser_msgbuf);					\
	}								\
} while (0)

#define PARSER_PHASE_FINAL	2
int		 parser_phase = 0;
int		 parser_debuglevel = 0;
int		 parser_not_initialized = 1;

#define PARSER_MSGBUFSIZE	256
static char	 parser_msgbuf[PARSER_MSGBUFSIZE];

static char	*aml_filename = NULL;

struct parser_state {
	int	 method_numargs;
	int	 pkglength;
	int	 pkglength_fixed;
	int	 buffersize;
	int	 buffersize_unknown;
	int	 in_method_interface;
	int	 prefix_expected;
};

struct parser_state_list {
	TAILQ_ENTRY(parser_state_list)	links;
	struct parser_state	*state;
};

TAILQ_HEAD(, parser_state_list)state_list, state_log;
struct parser_state	*current_state = NULL;
static void	 parser_state_push(void);
static void	 parser_state_pop(void);

static char	*parser_strcat(const char *str1, const char *str2);

static void	 parser_scope_enter(const char *name);
static void	 parser_scope_exit(void);
static char	*parser_scope_fullpath(const char *name);

static int	 parser_pkglength_encoding(int pkglength, u_int8_t *encoding);
static void	 parser_pkglength_fixup(void);

%}

%union {
	char			*str;
	u_int32_t		 num;
	union asl_literal	 lit;
	struct asl_symtab	*sym;
}

/*
 * Token
 */

%token		TOK_DEFINITIONBLOCK

%token		TOK_CREATEBYTEFIELD
%token		TOK_METHOD

%token		TOK_WHILE

%token		TOK_DECREMENT
%token		TOK_INDEX
%token		TOK_REFOF
%token		TOK_STORE

%token <num>	TOK_SERIALIZERULEKEYWORD

%token <num>	TOK_ARGTERM TOK_LOCALTERM
%token		TOK_DEBUGTERM

%token <lit>	TOK_BYTECONST TOK_WORDCONST TOK_DWORDCONST

%token <lit>	TOK_STRING

%token		TOK_BUFFER
%token		TOK_PACKAGE

%token <str>	TOK_NAMESEG

%token		TOK_ERROR

/*
 * Type
 */

%type <sym>	UserTerm

%type <sym>	CreateByteFieldTerm
%type <sym>	MethodTerm
%type <sym>	MethodInterface
%type <num>	MethodNumArgs
%type <num>	MethodSerializeRule

%type <num>	SerializeRuleKeyword

%type <lit>	ByteConst
%type <lit>	WordConst
%type <lit>	DWordConst

%type <lit>	String

%type <str>	NameSeg
%type <str>	NameString
%type <str>	PrefixPath
%type <str>	NamePath
%type <str>	NamePathTail

%%

ASLCode
	:	{ parser_state_push(); }
	  DefinitionBlockTerm  '{' ASLTermList '}'
		{
			PARSER_ENDPACKAGE("ASLTermList");
		}
	;

DefinitionBlockTerm
	: TOK_DEFINITIONBLOCK '(' String ',' String ',' ByteConst ',' String ',' String ',' DWordConst ')'
		{
			PARSER_DEBUGPRINT(1, "DefinitionBlock (");
			PARSER_DEBUGPRINT(1, "\"%s\", \"%s\", 0x%x, \"%s\", \"%s\", 0x%x)\n",
			    $3.str.ptr, $5.str.ptr,
			    $7.num.val, $9.str.ptr,
			    $11.str.ptr, $13.num.val);

			switch (parser_phase) {
			case 1:
				aml_filename = strdup($3.str.ptr);
				break;
			default:
				break;
			}

			free($3.str.ptr);
			free($5.str.ptr);
			free($9.str.ptr);
			free($11.str.ptr);
		}
	;

ASLTermList
	:	{ parser_state_push(); PARSER_BEGINPACKAGE("ASLTermList"); }
	  TermList
	;

TermList
	: Nothing | Term TermList
	;

Term
	: Object | Type1Opcode | Type2Opcode
	;

ObjectList
	: Nothing | Object ObjectList
	;

Object
	: /* CompilerDirective | */ NamedObject | /* NameSpaceModifier | */ UserTerm
	;

DataObject
	: BufferTerm | PackageTerm | LiteralData /* | DataMacros */
	;

LiteralData
	: Integer | String /* | ConstTerm */
	;

UserTerm
	: NameString '(' ArgList ')'
		{
			int	numargs;
			if (parser_phase == 2) {
				PARSER_SYMTAB_FIND($$, $1, "Method");
				if ($$ == NULL) {
					goto userterm_out;
				}
				numargs = symtab_method_getnumargs($$);
				if (numargs != current_state->method_numargs) {
					parser_error("number of args is wrong for %s\n", $1);
				}
			}
		userterm_out:
			current_state->method_numargs = 0;
			free($1);
		}
	| NameString	/* XXX Named object by user (not in Spec.) */
		{
			if (parser_phase == 2) {
				PARSER_SYMTAB_FIND($$,
				    parser_scope_fullpath($1), "Object");
			}
			free($1);
		}
	;

ArgList
	: Nothing
	| TermArg ArgListTail		{ current_state->method_numargs++; }
	;

ArgListTail
	: Nothing
	| ',' TermArg ArgListTail	{ current_state->method_numargs++; }
	;

TermArg
	: Type2Opcode | DataObject | UserTerm | ArgTerm | LocalTerm
	;

Type1Opcode
	: WhileTerm
	;

Type2Opcode
	: DecTerm | RefOfTerm | StoreTerm
	;

NamedObject
	: CreateByteFieldTerm | MethodTerm
	;

/*
 * Named Object
 */

CreateByteFieldTerm
	: CreateByteFieldDenotation '(' TermArg ',' TermArg ',' NameString ')'
		{
			if (parser_phase == 1) {
				PARSER_SYMTAB_NEW($$, parser_scope_fullpath($7), SYM_FIELD);
			}
			free($7);
		}
	;

CreateByteFieldDenotation
	: TOK_CREATEBYTEFIELD
		{
			PARSER_ENCODING_TERM(0x8c, "CreateByteField");
		}
	;

MethodTerm
	: MethodInterface '{' TermList '}'
		{
			$$ = $1;
			PARSER_ENDPACKAGE("Method");
			parser_scope_exit();
		}
	;

MethodInterface
	: MethodDenotation '(' NameString MethodNumArgs MethodSerializeRule ')'
		{
			current_state->in_method_interface = 0;
			PARSER_ENCODING_TERM(($4 | $5 << 3), "MethodFlags");
			if (parser_phase == 1) {
				PARSER_SYMTAB_NEW($$, parser_scope_fullpath($3), SYM_METHOD);
				symtab_method_setflags($$, (int)($4 | $5 << 3));
			}
			parser_scope_enter($3);
			free($3);
		}
	;

MethodDenotation
	: TOK_METHOD
		{
			PARSER_ENCODING_TERM(0x14, "Method");
			PARSER_BEGINPACKAGE("Method");
			parser_state_push();
			PARSER_PKGLENGTH_FLUSH();
			current_state->in_method_interface = 1;
		}
	;

MethodNumArgs
	: Nothing			{ $$ = 0; }
	| ',' ByteConst			{ $$ = $2.num.val; }
	;

MethodSerializeRule
	: Nothing			{ $$ = 0; }
	| ',' SerializeRuleKeyword	{ $$ = $2; }
	;

/*
 * Type1Opcode
 */

WhileTerm
	:  WhileDenotation '(' TermArg ')' '{' TermList '}'
		{
			PARSER_ENDPACKAGE("While");
		}
	;

WhileDenotation
	: TOK_WHILE
		{
			PARSER_ENCODING_TERM(0xa2, "While");
			PARSER_BEGINPACKAGE("While");
			parser_state_push();
			PARSER_PKGLENGTH_FLUSH();
		}
	;

/*
 * Type2Opcode
 */

DecTerm
	: DecrementDenotation '(' SuperName ')'
	;

DecrementDenotation
	: TOK_DECREMENT
		{
			PARSER_ENCODING_TERM(0x76, "Decrement");
		}
	;

IndexTerm
	: IndexDenotation '(' TermArg ',' TermArg IndexDestination ')'
	;

IndexDenotation
	: TOK_INDEX
		{
			PARSER_ENCODING_TERM(0x88, "Index");
		}
	;

IndexDestination
	: Nothing | ',' SuperName
	;


RefOfTerm
	: RefOfDenotation '(' SuperName ')'
	;

RefOfDenotation
	: TOK_REFOF
		{
			PARSER_ENCODING_TERM(0x71, "RefOf");
		}
	;

StoreTerm
	: StoreDenotation '(' TermArg ',' SuperName ')'
		{
			current_state->prefix_expected = 0;
		}
	;

StoreDenotation
	: TOK_STORE
		{
			PARSER_ENCODING_TERM(0x70, "Store");
			current_state->prefix_expected = 1;
		}
	;

/*
 * Keyword
 */

SerializeRuleKeyword
	: TOK_SERIALIZERULEKEYWORD	{ $$ = $1; }
	;

/*
 * SuperName
 */

SuperName
	: NameString | ArgTerm | LocalTerm | DebugTerm | IndexTerm
	;

ArgTerm
	: TOK_ARGTERM
		{
			PARSER_ENCODING_TERM(0x68 + $1, "Arg%d", $1);
		}
	;

LocalTerm
	: TOK_LOCALTERM
		{
			PARSER_ENCODING_TERM(0x60 + $1, "Local%d", $1);
		}
	;

DebugTerm
	: TOK_DEBUGTERM
	;

/*
 * Integer
 */

Integer
	: ByteConst | WordConst | DWordConst
	;

ByteConst
	: TOK_BYTECONST
		{
			$$ = $1;
			if (current_state->in_method_interface == 0) {
				PARSER_ENCODING_INTEGER($$.num.val,
				    1, "ByteConst");
			}
		}
	;

WordConst
	: TOK_WORDCONST
		{
			$$ = $1;
			PARSER_ENCODING_INTEGER($$.num.val, 2, "WordConst");
		}
	;

DWordConst
	: TOK_DWORDCONST
		{
			$$ = $1;
			PARSER_ENCODING_INTEGER($$.num.val, 4, "DWordConst");
		}
	;

/*
 * String
 */

String
	: TOK_STRING
		{
			char	*ptr;
			$$ = $1;
			PARSER_DEBUGPRINT(2, "\tString<%s>:\tat line %d\n",
			    $$.str.ptr, parser_lineno);
			if (current_state->prefix_expected) {
				PARSER_ENCODING_PREFIX(0, LITERAL_STRING);
			}
			ptr = $$.str.ptr;
			while (*ptr) {
				PARSER_ENCODING_TERM(*ptr, "String");
				ptr++;
			}
			PARSER_ENCODING_TERM(0x00, "StringTermination");
		}
	;

/*
 * Dataobject
 */

BufferTerm
	: BufferDenotation '(' BufferSize ')' '{' BufferBlock '}'
		{
			if (parser_phase == 1 && current_state->buffersize_unknown) {
				current_state->buffersize = current_state->pkglength;
				/* to adjust pkglength of the Buffer */
				current_state->prefix_expected = 1;
				PARSER_ENCODING_INTEGER(current_state->buffersize,
				    1, "BufferSize");	/* XXX */
				current_state->prefix_expected = 0;
			}

			PARSER_ENDPACKAGE("Buffer");
		}
	;

BufferDenotation
	: TOK_BUFFER
		{
			PARSER_ENCODING_TERM(0x11, "Buffer");
			PARSER_BEGINPACKAGE("Buffer");
			parser_state_push();
			PARSER_PKGLENGTH_FLUSH();
			if (current_state->buffersize) {
				current_state->prefix_expected = 1;
				PARSER_ENCODING_INTEGER(current_state->buffersize,
				    1, "BufferSize");	/* XXX */
				current_state->prefix_expected = 0;
			} else {
				current_state->prefix_expected = 1;
			}
		}
	;

BufferSize
	: Nothing
		{
			current_state->buffersize_unknown = 1;
			current_state->prefix_expected = 0;
		}
	| TermArg		{ current_state->prefix_expected = 0; }
	;

BufferBlock
	: String | ByteList
	;

ByteList
	: Nothing | ByteConst ByteListTail
	;

ByteListTail
	: Nothing | ',' ByteConst ByteListTail
	;

PackageTerm
	: PackageDenotation '(' PackageNumElements ')' '{' PackageList '}'
	;

PackageDenotation
	: TOK_PACKAGE
	;

PackageNumElements
	: Nothing | ByteConst
	;

PackageList
	: Nothing | PackageElement PackageListTail
	;

PackageListTail
	: Nothing | ',' PackageElement PackageListTail
	;

PackageElement
	: DataObject | NameString
	;

/*
 * ASL Names (15.1.2)
 */

RootChar
	: '\\'
	;

NameSeg
	: TOK_NAMESEG		{ $$ = $1; }
	;

NameString
	: RootChar NamePath
		{
			$$ = parser_strcat("\\", $2);
			free($2);
			goto namestring_out;
		}
	| PrefixPath NamePath
		{
			char	*ptr;
			$$ = parser_strcat($1, $2);
			free($1);
			free($2);
		namestring_out:
			PARSER_DEBUGPRINT(2, "\tNameString<%s>:\tat line %d\n",
			    $$, parser_lineno);
			ptr = $$;
			while (*ptr) {
				PARSER_ENCODING_TERM(*ptr, "NameString");
				ptr++;
			}
		}
	;

PrefixPath
	: Nothing		{ $$ = NULL; }
	| '^' PrefixPath	{ $$ = parser_strcat("^", $2); free($2); }
	;

NamePath
	: Nothing		{ $$ = NULL; }
	| NameSeg NamePathTail	{ $$ = parser_strcat($1, $2); free($1); free($2); }
	;

NamePathTail
	: Nothing		{ $$ = NULL; }
	| '.' NameSeg NamePathTail
		{
			$$ = parser_strcat(".", parser_strcat($2, $3));
			free($2);
			free($3);
		}
	;

/*
 * Nothing
 */

Nothing
	: /* empty */
	;

%%

static char *
parser_strcat(const char *str1, const char *str2)
{
	char	*str;
	size_t	len;

	len = ((str1 == NULL) ? 0 : strlen(str1)) +
	      ((str2 == NULL) ? 0 : strlen(str2)) + 1;

	if ((str = (char *)malloc(len)) == NULL) {
		parser_fatal("No more room for strcat");
	}

	bzero(str, len);
	if (str1 != NULL) {
		strncpy(str, str1, strlen(str1));
	}
	if (str2 != NULL) {
		strcat(str, str2);
	}

	return(str);
}

static int
parser_pkglength_encoding(int pkglength, u_int8_t *encoding)
{
	int		len;
	u_int8_t	code[4];

	len = 0;
	bzero(code, sizeof(code));
	if (pkglength <= 0x0000003f) {	/* 6bit */
		len = 1;
		code[0] = pkglength;
		goto out;
	}
	if (pkglength <= 0x00000fff) {	/* 12bit */
		len = 2;
		code[0] = (1 << 6) + (pkglength & 0xf);
		code[1] = (pkglength >> 4) & 0xff;
		goto out;
	}
	if (pkglength <= 0x000fffff) {	/* 20bit */
		len = 3;
		code[0] = (2 << 6) + (pkglength & 0xf);
		code[1] = (pkglength >> 4) & 0xff;
		code[2] = (pkglength >> 12) & 0xff;
		goto out;
	}
	if (pkglength <= 0x0fffffff) {	/* 28bit */
		len = 4;
		code[0] = (3 << 6) + (pkglength & 0xf);
		code[1] = (pkglength >> 4) & 0xff;
		code[2] = (pkglength >> 12) & 0xff;
		code[3] = (pkglength >> 20) & 0xff;
		goto out;
	}
out:
	if (encoding != NULL) {
		bcopy(code, encoding, sizeof(code));
	}
	return(len);
}

static void
parser_pkglength_fixup(void)
{
	int		pkglength;

	if (current_state->pkglength_fixed == 0) {
		pkglength = current_state->pkglength;
		pkglength += parser_pkglength_encoding(pkglength, NULL);
		current_state->pkglength = pkglength;
		current_state->pkglength_fixed = 1;
	}
}

static __inline void
parser_state_init()
{
	static int	first = 1;

	if (first) {
		TAILQ_INIT(&state_list);
		TAILQ_INIT(&state_log);
		first = 0;
	}
}

static void
parser_state_push(void)
{
	struct		parser_state *new_state;
	struct		parser_state_list *state_entry, *entry_log;

	parser_state_init();
	switch (parser_phase) {
	case 1:
		new_state = (struct parser_state *)malloc(sizeof(struct parser_state));
		if (new_state == NULL) {
			parser_fatal("No room for parser_state.");
		}
		bzero(new_state, sizeof(struct parser_state));
		state_entry = (struct parser_state_list *)malloc(sizeof(struct parser_state_list));
		if (state_entry == NULL) {
			parser_fatal("No room for parser_state_list.");
		}
		entry_log = (struct parser_state_list *)malloc(sizeof(struct parser_state_list));
		if (entry_log == NULL) {
			parser_fatal("No room for parser_state_list.");
		}

		state_entry->state = entry_log->state = new_state;

		TAILQ_INSERT_TAIL(&state_log, entry_log, links);
		break;
	default:
		state_entry = TAILQ_FIRST(&state_log);
		TAILQ_REMOVE(&state_log, state_entry, links);
		TAILQ_INSERT_TAIL(&state_log, state_entry, links);
		break;
	}
	TAILQ_INSERT_HEAD(&state_list, state_entry, links);
	current_state = state_entry->state;
}

static void
parser_state_pop(void)
{
	int		pkglength;
	struct		parser_state_list *state_entry;

	parser_state_init();
	pkglength = current_state->pkglength;
	state_entry = TAILQ_FIRST(&state_list);
	TAILQ_REMOVE(&state_list, state_entry, links);
	if (parser_phase == 1) {
		free(state_entry);
	}

	state_entry = TAILQ_FIRST(&state_list);
	current_state = state_entry->state;
	if (parser_phase == 1) {
		current_state->pkglength += pkglength;
	}
}

struct parser_scope {
	TAILQ_ENTRY(parser_scope)	links;
	char			*name;
};
TAILQ_HEAD(, parser_scope)	scope_list;

static __inline void
parser_scope_init()
{
	static int	first = 1;

	if (first) {
		TAILQ_INIT(&scope_list);
		first = 0;
	}
}

static void
parser_scope_enter(const char *name)
{
	struct parser_scope *new_entry;

	parser_scope_init();

	if ((new_entry = malloc(sizeof(struct parser_scope))) == NULL) {
		parser_fatal("No more room for parser_scope.");
	}

	bzero(new_entry, sizeof(struct parser_scope));
	new_entry->name = strdup(name);
	TAILQ_INSERT_HEAD(&scope_list, new_entry, links);
}

static void
parser_scope_exit(void)
{
	struct parser_scope *entry;

	parser_scope_init();
	entry = TAILQ_FIRST(&scope_list);
	TAILQ_REMOVE(&scope_list, entry, links);
	free(entry->name);
	free(entry);
}

static char *
parser_scope_fullpath(const char *name)
{
	struct parser_scope *entry;
	static char	fullpath[PARSER_MSGBUFSIZE];

	parser_scope_init();
	bzero(fullpath, sizeof(fullpath));
	TAILQ_FOREACH(entry, &scope_list, links) {
		snprintf(fullpath, sizeof(fullpath), "%s%s.", fullpath, entry->name);
	}
	snprintf(fullpath, sizeof(fullpath), "%s%s", fullpath, name);
	return(fullpath);
}

void
yyerror(const char *msg)
{

	fprintf(stderr, "line %d: %s\n", parser_lineno, msg);
}

#define PARSER_WARN(msg)	do {					\
	va_list	ap;							\
	va_start(ap, msg);						\
	vsnprintf(parser_msgbuf, PARSER_MSGBUFSIZE, msg, ap);		\
	va_end(ap);							\
	yyerror(parser_msgbuf);						\
} while (0)

void
parser_warn(const char *msg, ...)
{
	PARSER_WARN(msg);
}

void
parser_error(const char *msg, ...)
{
	PARSER_WARN(msg);
	exit(1);
}

void
parser_fatal(const char *msg, ...)
{
	PARSER_WARN(msg);
	abort();
}

extern FILE	*yyin;

void
parser_main(const char *asl_filename)
{

	if ((yyin = fopen(asl_filename, "r")) == NULL) {
		(void) err(1, "cannot open ASL file");
	}

	for (parser_phase = 1; parser_phase <= PARSER_PHASE_FINAL; parser_phase++) {
		parser_lineno = 1;
		if (parser_debuglevel) {
			printf("==== PARSER PHASE %d ====\n", parser_phase);
		}
		
		if (yyparse() != 0) {
			fprintf(stderr, "ASL parser errors in phase %d\n", parser_phase);
			exit(1);
		}

		switch (parser_phase) {
		case 1:
			symtab_dump();
			break;
		default:
			break;
		}

		rewind(yyin);
	}

	fclose(yyin);
	symtab_destroy();
}
